local M = {}

local helper = require('scenario/scenariohelper')

local max_delivery_distance = 1500
local pizza_shop_pos = vec3(1053.89319, 1323.19946, 145.01397)

local deliver_pizza_msgs = {"The pizza is getting cold Mario!", "What are you waiting for Mario? Get moving!", "We don't have all day Mario!"}
local thanks_msgs = {"Grazie!", "What took you so long!", "Keep the change!", "Why is my pizza covered in soda!", "Where's the pineapple?!", "Thanks for nothing!"}

local buildings_for_delivery = {}

local state = "scenario_started"

local the_building_pos = vec3()
local destination_pos = vec3()

local the_wp1_pos = vec3()
local the_wp2_pos = vec3()

local deliveries_made = 0

local distance_to_destination = 0
local current_delivery_distance = 0

local time_left = 1
local countdown_display_timer = 1
local update_ui_timer = 0
local update_ui_delay = 0.25 -- update every 0.25 seconds

-- http://lua-users.org/wiki/SimpleRound
function round(num, numDecimalPlaces)
  local mult = 10^(numDecimalPlaces or 0)
  return math.floor(num * mult + 0.5) / mult
end

-- Edited function from groundMarkers.lua since it doesn't work...
local function getPathLength(path)
  local dist = 0
  local pathLength = tableSize(path)
  for i = 1, (pathLength - 1) do
    dist = dist + (path[i + 1].pos - path[i].pos):length()
  end
  return dist
end

local function getRandomBuilding()
	return buildings_for_delivery[math.random(#buildings_for_delivery)]
end

local function getClosestRoadPos(building_pos)
	local wp1, wp2, lat_dist_from_wp = map.findClosestRoad(building_pos)
	
	local wp1_pos = vec3(map.getMap().nodes[wp1].pos)
	local wp2_pos = vec3(map.getMap().nodes[wp2].pos)
	
	local xnorm = clamp(building_pos:xnormOnLine(wp1_pos, wp2_pos), 0, 1)
  local wp_dir = wp2_pos - wp1_pos
  local closest_pos_on_road_to_building = xnorm * wp_dir + wp1_pos
	
	local road_width = map.getMap().nodes[wp1].radius
	
	local offset_vec = (road_width - 1.5) * (building_pos - closest_pos_on_road_to_building):normalized()
	
	return closest_pos_on_road_to_building + offset_vec, wp1, wp2, wp_dir
end

local function getDeliveryTime(dist)
	local avg_speed = (1.4 + deliveries_made * 0.1) * math.pow(dist, 1 / 3) 
	
	return dist / avg_speed
end

local function setDestination(pos, trigger_object)
	local wp1, wp2, wp_dir = nil
	
	destination_pos, wp1, wp2, wp_dir = getClosestRoadPos(pos)
	
	the_wp1_pos = vec3(map.getMap().nodes[wp1].pos)
	the_wp2_pos = vec3(map.getMap().nodes[wp2].pos)
	
	local rot = quatFromDir(wp_dir, vec3(0,0,1))
	
	if trigger_object then
		trigger_object:setPosRot(destination_pos.x, destination_pos.y, destination_pos.z, rot.x, rot.y, rot.z, rot.w)
	end
	
	core_groundMarkers.setFocus(wp1)
	current_delivery_distance = core_groundMarkers.getPathLength()
	
	--core_groundMarkers.setFocus(wp1, math.huge) -- Sets navigation to waypoint and removes ground arrow markers
	--core_groundMarkers.onSerialize() -- Gets rid of floating directional arrows
end

local function goDeliverPizza()
	local delivery_point_trigger = scenetree.findObject("delivery_point_trigger")
	
	-- Set destination to random building
	local rand_building = getRandomBuilding()
	the_building_pos = vec3(rand_building:getPosition())
	setDestination(the_building_pos, delivery_point_trigger)

	--print("Delivery distance = " .. current_delivery_distance)

	local delivery_time = getDeliveryTime(current_delivery_distance)
	time_left = time_left + delivery_time	
	
	local rand_msg = deliver_pizza_msgs[math.random(#deliver_pizza_msgs)]
	
	helper.flashUiMessage(rand_msg .. " The place is " .. string.format("%.2f", current_delivery_distance / 1000.0) .. "km away and you have " .. tostring(math.ceil(time_left)) .. " seconds to get there!", 7)
	
	state = "deliver_pizza"
end

local function goBackToPizzaShop()
	setDestination(pizza_shop_pos)
	
	--print("Back to shop distance = " .. current_delivery_distance)
	local delivery_time = getDeliveryTime(current_delivery_distance)
	time_left = time_left + delivery_time
	
	local rand_msg = thanks_msgs[math.random(#thanks_msgs)]
	
	helper.flashUiMessage(rand_msg .. " You have " .. tostring(math.ceil(time_left)) .. " seconds to go back!", 5)
	
	deliveries_made = deliveries_made + 1
	
	state = "retrieve_pizza"
end

local function displayUIText(dt)
	if update_ui_timer >= update_ui_delay then
		local rounded_dist = string.format("%.2f", core_groundMarkers.getPathLength() / 1000.0, 2)
		
		if state == "deliver_pizza" then
			guihooks.message("Deliver the pizza (" .. tostring(rounded_dist) .. "km)", 2 * update_ui_delay, "delivery_status")
		elseif state == "retrieve_pizza" then
			guihooks.message("Go back to pizza shop (" .. tostring(rounded_dist) .. "km)", 2 * update_ui_delay, "delivery_status")
		end
		
		guihooks.message("Time Left: " .. tostring(math.ceil(time_left)) .. "s", 2 * update_ui_delay, "time_left", "timer")
		guihooks.message("Deliveries: " .. tostring(deliveries_made), 2 * update_ui_delay, "num_deliveries", "local_pizza")
	
		update_ui_timer = 0
	end
	
	update_ui_timer = update_ui_timer + dt
end

local function onPreRender(dt, dtSim)
	local veh = be:getPlayerVehicle(0)
	if not veh or state == "scenario_started" or not be:getEnabled() then return end

	-- Display UI text
	displayUIText(dt)
	
	-- Cylinder to indicate destination
	debugDrawer:drawCylinder((destination_pos + vec3(0, 0, -5)):toPoint3F(), (destination_pos + vec3(0, 0, 75)):toPoint3F(), 1, ColorF(1,0,0,0.25))
	
	if time_left <= 0 then
		-- Game over
		if state == "deliver_pizza" then
			scenario_scenarios.finish({failed = "The pizza became inedible! You delivered a total of " .. tostring(deliveries_made) .. " pizzas."})
		elseif state == "retrieve_pizza" then
			scenario_scenarios.finish({failed = "You didn't make it back on time! You delivered a total of " .. tostring(deliveries_made) .. " pizzas."})
		end
		
		state = "scenario_started"
		
	elseif time_left <= 10 then
		-- When 10 seconds are left: every second display countdown timer
		
		if countdown_display_timer >= 1 then
			helper.flashUiMessage(tostring(math.ceil(time_left)), 0.5, true)	
			
			countdown_display_timer = 0
		end
		countdown_display_timer = countdown_display_timer + dt
		
	else
		countdown_display_timer = 1
		
	end
	
	time_left = time_left - dt
	
	--debugDrawer:drawCylinder((the_building_pos):toPoint3F(), (the_building_pos + vec3(0, 0, 20)):toPoint3F(), 1, ColorF(0,1,0,1))
	--debugDrawer:drawCylinder((the_wp1_pos):toPoint3F(), (the_wp1_pos + vec3(0, 0, 5)):toPoint3F(), 1, ColorF(1,0,0,1))
	--debugDrawer:drawCylinder((the_wp2_pos):toPoint3F(), (the_wp2_pos + vec3(0, 0, 5)):toPoint3F(), 1, ColorF(1,0,0,1))
end

local function onBeamNGTrigger(data)
	if data.subjectName ~= "thePlayer" then return end -- Only care about player on triggers
	
	local stopped = vec3(be:getPlayerVehicle(0):getVelocity()):length() < 0.5
	
	if stopped then
		if data.triggerName == "delivery_point_trigger" then
			if state == "deliver_pizza" then
				Engine.Audio.playOnce('AudioGui', "event:UI_Checkpoint")
				goBackToPizzaShop()
			end
		elseif data.triggerName == "pizza_shop_trigger" then
			if state == "retrieve_pizza" then		
				Engine.Audio.playOnce('AudioGui', "event:UI_Checkpoint")
				goDeliverPizza()
			end
		end
	end
end

local function getDistancesToBuildings(buildings)
	local arr = {}
	
	local player_pos = vec3(be:getPlayerVehicle(0):getPosition())
	
	for key, object in pairs(buildings) do
		local building_pos = vec3(object:getPosition())
		
		local distance = (player_pos - building_pos):length()
		
		table.insert(arr, {object = object, distance = distance})
	end
	
	--sortDistances(arr)
	
	return arr
end

local function onScenarioLoaded(scenario)
	--local buildings = Sim.findObject('BUILDINGS'):getObjects()
	local objects = getObjectsByClass('TSStatic')

	local datas = getDistancesToBuildings(objects)
	
	for key, data in pairs(datas) do
		local object = data.object
		local distance = data.distance
		
		-- Filtering out bus stops/signs and bridges
		if distance < max_delivery_distance and object.shapeName:match("levels/italy/art/shapes/buildings") then	
			if object.shapeName:match("house") 
			or object.shapeName:match("town_bld") 
			then
				table.insert(buildings_for_delivery, object)
			end
		end
	end
	
	--[[
	for key, object in pairs(objects) do
		-- Filtering out bus stops/signs and bridges
		if object.shapeName:match("levels/italy/art/shapes/buildings") then	
			if object.shapeName:match("house") 
			or object.shapeName:match("town_bld") 
			then
				table.insert(buildings_for_delivery, object)
			end
		end
	end
	]]--
	
	--gameplay_traffic.setupTraffic(getMaxVehicleAmount())
	--gameplay_traffic.setTrafficVars({spawnValue = 1.3, spawnDirBias = 0.3})
end
	
local function onRaceStart()
	goDeliverPizza()
end

local function onScenarioRestarted(scenario)
  core_groundMarkers.resetAll()
	countdown_display_timer = 1
	time_left = 1
	deliveries_made = 0
	
	state = "scenario_started"
end

M.onPreRender = onPreRender
M.onBeamNGTrigger = onBeamNGTrigger
M.onScenarioLoaded = onScenarioLoaded
M.onRaceStart = onRaceStart
M.onScenarioRestarted = onScenarioRestarted

return M